# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/95_cli.ipynb (unless otherwise specified).

__all__ = ['logger', 'get_parser', 'main']

# Cell

import argparse
import logging
from pathlib import Path
import random

from .utils import log_elapsed_time
from .coco import merge_datasets, cut_annotations_per_category, remove_invalid_elements
from .json_file import *
from .json_tree import *
from .crop_tree import *

# Cell
logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
logger = logging.getLogger()

# Cell

def get_parser():
    parser = argparse.ArgumentParser(
        description="Tool for converting datasets in COCO format between different representations"
    )
    parser.add_argument("--in_json_file", type=Path, nargs="*", default=[],
                        help=(
                            "Path to one or multiple json files storing COCO dataset "
                            "in `json_file` representation (all json-based datasets will be merged)."
                        ))

    parser.add_argument("--in_json_tree", type=Path, nargs="*", default=[],
                        help=(
                            "Path to one or multiple directories storing COCO dataset "
                            "in `json_tree` representation (all json-based datasets will be merged)."
                        ))

    parser.add_argument("--in_crop_tree", type=Path, nargs="*", default=[],
                        help=(
                            "Path to one or multiple directories storing COCO dataset "
                            "in `crop_tree` representation (all crop-based datasets will be merged and will "
                            "overwrite the json-based datasets)."
                       ))

    parser.add_argument("--out_path", type=Path,
                        help="Path to the output dataset (file or directory: depends on `--out_format`)")

    parser.add_argument("--out_format", choices=['json_file', 'json_tree', 'crop_tree'])

    parser.add_argument("--seed", type=int, default=42, help="Random seed.")

    parser.add_argument("--max_crops_per_class", type=int, default=None,
                        help=(
                            "If set, the tool will randomly select up to this number of "
                            "crops (annotations) per each class (category) and drop the others."),
                       )

    parser.add_argument("--drop_invalid_elements", action='store_true',
                        help="If set, drops broken elements (for example, negative IDs or broken bboxes).")

    parser.add_argument("--dump_crop_tree_num_processes", type=int, default=1)

    parser.add_argument("--overwrite", action='store_true',
                        help="If set, will delete the output file/directory before dumping the result dataset.")
    parser.add_argument("--indent", default=4,
                        type=lambda x: int(x) if str(x).lower() not in ('none', 'null', '~') else None,
                        help="Indentation in the output json files.")

    parser.add_argument("--debug", action='store_true')

    return parser

# Cell

@log_elapsed_time(lambda t: logger.info(f'Total elapsed: {t.elapsed}'))
def main(args=None):
    args = args or get_parser().parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    logger.info(f'Arguments: {args}')

    in_json_tree_list = args.in_json_tree
    in_json_file_list = args.in_json_file
    in_crop_tree_list = args.in_crop_tree

    seed = args.seed
    max_crops_per_class = args.max_crops_per_class
    drop_invalid_elements = args.drop_invalid_elements

    out_path = args.out_path
    out_format = args.out_format
    dump_crop_tree_num_processes = args.dump_crop_tree_num_processes
    overwrite = args.overwrite
    indent = args.indent

    if out_path and not out_format or not out_path and out_format:
        raise ValueError(f'Option --out_format requires --out_path and vice versa')

    random.seed(args.seed)

    coco = None
    coco_count = 0
    for in_json_tree in in_json_tree_list:
        coco = merge_datasets(coco, load_json_tree(in_json_tree))
        coco_count += 1
    for in_json_file in in_json_file_list:
        coco = merge_datasets(coco, load_json_file(in_json_file))
        coco_count += 1

    if coco is None:
        raise ValueError(f'Not found base dataset, please specify either of: '
                         '--in_json_tree / --in_json_file (multiple arguments allowed)')
    if coco_count > 1:
        logger.info(f'Total loaded json dataset: {coco.to_full_str()}')

    coco_crop = None
    coco_crop_count = 0
    for in_crop_tree in in_crop_tree_list:
        coco_crop = merge_datasets(coco_crop, load_crop_tree(in_crop_tree, coco))
        coco_crop_count += 1
    if coco_crop is not None:
        if coco_crop_count > 1:
            logger.info(f'Total loaded crop-tree dataset: {coco_crop.to_full_str()}')
        logger.info('Using coco_crop dataset only.')
        coco = coco_crop

    if max_crops_per_class:
        logger.info(f'Cutting off crops up to {max_crops_per_class} per class, random seed={seed}')
        coco = cut_annotations_per_category(coco, max_crops_per_class)
        logger.info(f'After cutting off: {coco.to_full_str()}')

    if drop_invalid_elements:
        coco = remove_invalid_elements(coco)
        logger.info(f'After removing invalid elements: {coco.to_full_str()}')

    logger.info(f'Result dataset: {coco.to_full_str()}')
    details = ''
    if out_format is not None:
        assert out_path
        dump_kwargs = dict(skip_nulls=True, overwrite=overwrite, indent=indent)
        if out_format == 'json_file':
            dump_fun = dump_json_file
        elif out_format == 'json_tree':
            dump_fun = dump_json_tree
        elif out_format == 'crop_tree':
            dump_fun = dump_crop_tree
            dump_kwargs['num_processes'] = dump_crop_tree_num_processes
        else:
            raise ValueError(out_format)
        dump_fun(coco, out_path, **dump_kwargs)
        if out_path.is_dir():
            details = f': {[p.name for p in out_path.iterdir()]}'

    logger.info(f'[+] Success: {out_format} dumped to {out_path}' + details)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_json_file.ipynb (unless otherwise specified).

__all__ = ['logger', 'load_json_file', 'dump_json_file']

# Cell

import logging
import time
from dataclasses import dataclass
from typing import *
from pathlib import Path
import json

from .utils import sort_dict, measure_time
from .coco import *


# Cell
logger = logging.getLogger()

# Cell

def load_json_file(annotations_json: Union[str, Path], *, kind: str = "object_detection") -> CocoDataset:
    from_dict_function = get_dataset_class(kind).from_dict

    annotations_json = Path(annotations_json)
    logger.info(f"Loading json_file from: {annotations_json}")
    ext = annotations_json.suffix
    if ext != '.json':
        raise ValueError(f'Expect .json file as input, got: {annotations_json}')

    with measure_time() as timer:

        with measure_time() as timer2:
            D = json.loads(annotations_json.read_text())
        logger.info(f"  json file loaded: elapsed {timer2.elapsed}")

        with measure_time() as timer2:
            coco = from_dict_function(D)
        logger.info(f"  dataset constructed: elapsed {timer2.elapsed}")

    logger.info(f"Loaded json_file: elapsed {timer.elapsed}: {coco.to_full_str()}")

    return coco

# Cell

def dump_json_file(
    coco: CocoDataset,
    annotations_json: Union[str, Path],
    *,
    kind: str = "object_detection",
    skip_nulls: bool = True,
    overwrite: bool = False,
    indent: Optional[int] = 4,
) -> None:
    dataset_class = get_dataset_class(kind)
    if skip_nulls:
        to_dict_function = dataset_class.to_dict_skip_nulls
    else:
        to_dict_function = dataset_class.to_dict

    annotations_json = Path(annotations_json)
    if annotations_json.is_file() and not overwrite:
        raise ValueError(f"Destination json_file already exists: {annotations_json}")

    with measure_time() as timer:
        raw = sort_dict(to_dict_function(coco))
    logger.info(f"Sorted keys: elapsed {timer.elapsed}")

    logger.info(f"Writing dataset {coco.to_full_str()} to json-file: {annotations_json}")
    with measure_time() as timer:
        annotations_json.parent.mkdir(parents=True, exist_ok=True)
        annotations_json.write_text(json.dumps(raw, indent=indent, ensure_ascii=False))
    logger.info(f"Dataset written to {annotations_json}: elapsed {timer.elapsed}")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_json_tree.ipynb (unless otherwise specified).

__all__ = ['logger', 'load_json_tree', 'dump_json_tree']

# Cell

import logging
from typing import *
from pathlib import Path
import json
import shutil

from .utils import sort_dict, measure_time
from .coco import *

# Cell
logger = logging.getLogger()

# Cell

def load_json_tree(tree_dir: Union[str, Path], *, kind: str = "object_detection") -> CocoDataset:
    dataset_class = get_dataset_class(kind)
    from_dict_function = dataset_class.from_dict

    tree_dir = Path(tree_dir)
    logger.info(f"Loading json_tree from dir: {tree_dir}")
    if not tree_dir.is_dir():
        raise ValueError(f"Source json_tree dir not found: {tree_dir}")

    D = {}
    with measure_time() as timer:

        with measure_time() as timer2:
            for el_name in dataset_class.get_collective_elements():
                el_dir = tree_dir / el_name
                if not el_dir.is_dir():
                    logger.debug(f'Chunks dir not found: {el_dir}')
                    el_list = []
                else:
                    el_list = [json.loads(f.read_text()) for f in el_dir.glob('*.json')]
                logger.debug(f'Loaded {len(el_list)} json chunks from {el_dir}')
                D[el_name] = el_list
            for el_name in dataset_class.get_non_collective_elements():
                el_file = tree_dir / f'{el_name}.json'
                if not el_file.is_file():
                    logger.debug(f'Chunks file not found: {el_file}')
                    el = {}
                else:
                    el = json.loads(el_file.read_text())
                logger.debug(f'Loaded single-file json chunk {el_file}')
                D[el_name] = el
        logger.info(f"  json files loaded: elapsed {timer2.elapsed}")

        with measure_time() as timer2:
            coco = dataset_class.from_dict(D)
        logger.info(f"  dataset constructed: elapsed {timer2.elapsed}")

    logger.info(f"Loaded from json_tree: {coco.to_full_str()}")
    return coco

# Cell

def dump_json_tree(
    coco: CocoDataset,
    target_dir: Union[str, Path],
    *,
    kind: str = 'object_detection',
    skip_nulls: bool = True,
    overwrite: bool = False,
    indent: Optional[int] = 4,
) -> None:
    dataset_class = get_dataset_class(kind)
    if skip_nulls:
        to_dict_function = dataset_class.to_dict_skip_nulls
    else:
        to_dict_function = dataset_class.to_dict

    target_dir = Path(target_dir)
    raw = to_dict_function(coco)
    logger.info(f"Dumping json_tree to dir: {target_dir}")

    if overwrite:
        if target_dir.is_dir():
            logger.warning(f'Destination dir exists and will be overwritten: {target_dir}')
    elif target_dir.is_dir():
        raise ValueError(f"Destination json tree dir already exists: {target_dir}")

    if target_dir.is_dir():
        logger.info(f'Deleting old target tree directory {target_dir}')
        shutil.rmtree(str(target_dir))

    target_dir.mkdir(parents=True)

    with measure_time() as timer:
        # TODO: rename cat -> el_kind
        for cat in dataset_class.get_collective_elements():
            el_dir = target_dir / cat
            if not raw.get(cat):
                logger.debug(f'Skipping empty category {el_dir}')
                continue
            el_dir.mkdir()
            for el in raw[cat]:
                el_file = el_dir / f'{el["id"]}.json'
                el = sort_dict(el)
                el_file.write_text(json.dumps(el, indent=indent, ensure_ascii=False))
            logger.debug(f'Written {len(raw[cat])} elements to {el_dir}')

        for cat in dataset_class.get_non_collective_elements():
            el_dir = target_dir / cat
            el_dir.mkdir()
            el_file = target_dir / f'{cat}.json'
            el = raw[cat]
            el = sort_dict(el)
            el_file.write_text(json.dumps(el, indent=indent, ensure_ascii=False))
            logger.debug(f'Written single element to {el_dir}')
    logger.info(f'Dataset written to {target_dir}: elapsed {timer.elapsed}')


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"logger": "95_cli.ipynb",
         "CocoElement": "01_coco.ipynb",
         "CocoInfo": "01_coco.ipynb",
         "CocoLicense": "01_coco.ipynb",
         "CocoImage": "01_coco.ipynb",
         "CocoAnnotation": "01_coco.ipynb",
         "CocoObjectDetectionAnnotation": "01_coco.ipynb",
         "CocoCategory": "01_coco.ipynb",
         "CocoObjectDetectionCategory": "01_coco.ipynb",
         "CocoDataset": "01_coco.ipynb",
         "CocoObjectDetectionDataset": "01_coco.ipynb",
         "get_dataset_class": "01_coco.ipynb",
         "MAP_COCO_TYPE_TO_DATASET_CLASS": "01_coco.ipynb",
         "merge_datasets": "01_coco.ipynb",
         "shuffle": "01_coco.ipynb",
         "cut_annotations_per_category": "01_coco.ipynb",
         "remove_invalid_elements": "01_coco.ipynb",
         "load_json_file": "02_json_file.ipynb",
         "dump_json_file": "02_json_file.ipynb",
         "load_json_tree": "03_json_tree.ipynb",
         "dump_json_tree": "03_json_tree.ipynb",
         "load_crop_tree": "04_crop_tree.ipynb",
         "dump_crop_tree": "04_crop_tree.ipynb",
         "TimerMutable": "90_utils.ipynb",
         "measure_time": "90_utils.ipynb",
         "log_elapsed_time": "90_utils.ipynb",
         "sort_dict": "90_utils.ipynb",
         "sanitize_filename": "90_utils.ipynb",
         "read_image": "90_utils.ipynb",
         "download_image": "90_utils.ipynb",
         "cut_bbox": "90_utils.ipynb",
         "write_image": "90_utils.ipynb",
         "get_parser": "95_cli.ipynb",
         "main": "95_cli.ipynb"}

modules = ["coco.py",
           "json_file.py",
           "json_tree.py",
           "crop_tree.py",
           "utils.py",
           "cli.py"]

doc_url = "https://neuro-inc.github.io/cocorepr/cocorepr/"

git_url = "https://github.com/neuro-inc/cocorepr"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/90_utils.ipynb (unless otherwise specified).

__all__ = ['TimerMutable', 'measure_time', 'log_elapsed_time', 'sort_dict', 'sanitize_filename', 'read_image',
           'download_image', 'cut_bbox', 'write_image']

# Cell

import re
import cv2
import datetime
import logging
import urllib.request
import json
import time
from dataclasses import dataclass
from contextlib import contextmanager
from collections import OrderedDict
from pathlib import Path
from typing import *


# Cell

@dataclass(frozen=False)
class TimerMutable:
    elapsed: Optional[datetime.timedelta] = None


@contextmanager
def measure_time():
    timer = TimerMutable()
    start = time.time()
    yield timer
    timer.elapsed = datetime.timedelta(seconds=time.time() - start)

# Cell

def log_elapsed_time(log_method):
    def log(func):
        def wrapped(*args, **kwargs):
            timer = TimerMutable()
            try:
                with measure_time() as timer:
                    func(*args, **kwargs)
            finally:
                log_method(timer)

        return wrapped
    return log

# Cell

def sort_dict(D: Dict, sort_key='id') -> Dict:
    assert isinstance(D, dict), (type(D), D)
    return OrderedDict({
        k: (
            sorted(D[k], key=lambda x: x[sort_key])
            if isinstance(D[k], list) \
                and D[k] \
                and isinstance(D[k][0], dict) \
                and (sort_key in D[k][0])
            else D[k]
        )
        for k in sorted(D.keys())
    })

# Cell

def sanitize_filename(s: str, max_len=256) -> str:
    s = re.sub(r'[-.\s]+', '_', s)
    s = re.sub(r'[^\w ]', '', s)
    s = s[:max_len] if len(s) > max_len else s
    s = s.strip('_')
    return s

# Cell

def read_image(
    image_path: Union[str, Path],
    download_url: Optional[str] = None,
) -> "np.ndarray":
    """ Reads image :image_path: in RGB mode. If image file does not
        exist and :download_url: was specified, downloads the image first.
    """
    if download_url:
        download_image(image_path, download_url)

    assert Path(image_path).is_file(), f'Image not exists: {image_path}'

    image = cv2.imread(str(image_path))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image

# Cell

def download_image(image_path: Union[str, Path], download_url: str):
    image_path = Path(image_path)
    if image_path.exists():
        return
    image_path.parent.mkdir(exist_ok=True, parents=True)
    urllib.request.urlretrieve(download_url, str(image_path))

# Cell

def cut_bbox(image, bbox):
    x, y, w, h = map(int, bbox)
    crop = image[y:(y+h), x:(x+w)]
    return crop

# Cell

def write_image(image, image_path):
    Path(image_path).parent.mkdir(exist_ok=True, parents=True)
    try:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        cv2.imwrite(str(image_path), image)
    except cv2.error as e:
        raise ValueError(f'Could not write image {image_path}: {e}')


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_coco.ipynb (unless otherwise specified).

__all__ = ['logger', 'CocoElement', 'CocoInfo', 'CocoLicense', 'CocoImage', 'CocoAnnotation',
           'CocoObjectDetectionAnnotation', 'CocoCategory', 'CocoObjectDetectionCategory', 'CocoDataset',
           'CocoObjectDetectionDataset', 'get_dataset_class', 'MAP_COCO_TYPE_TO_DATASET_CLASS', 'merge_datasets',
           'shuffle', 'cut_annotations_per_category', 'remove_invalid_elements']

# Cell

import logging
import random
from abc import abstractmethod
from datetime import datetime
from dataclasses_json import dataclass_json
from dataclasses import dataclass, fields, asdict, field, replace
from typing import *
from pathlib import Path

from .utils import sanitize_filename

# Type helpers:
X, Y, W, H = Type[int], int, int, int

# Cell
logger = logging.getLogger()

# Cell

@dataclass_json
@dataclass
class CocoElement:
    def to_dict_skip_nulls(self):
        """ Same as `self.to_dict()` but does not add those fields
            whose values are `None`.
            WARNING! If you explicitly set a `None` value to a field
                     that has a non-`None` default value, it still
                     won't be dumped and will be deserialized wrongly.
        """
        return asdict(
            self,
            dict_factory=(
                lambda kv: {
                    k: v
                    for k, v in kv
                    if v is not None
                }
            )
        )

    @property
    def collection_name(self) -> str:
        raise NotImplementedError

    def is_valid(self) -> bool:
        raise NotImplementedError

# Cell

@dataclass
class CocoInfo(CocoElement):
    year: Optional[int] = None
    version: Optional[str] = None
    description: Optional[str] = None
    contributor: Optional[str] = None
    url: Optional[str] = None
    date_created: Optional[str] = None

    @property
    def collection_name(self):
        return "info"

    def is_valid(self) -> bool:
        return True  # no restrictions on the format

# Cell

@dataclass
class CocoLicense(CocoElement):
    id: int
    name: str
    url: Optional[str] = None

    @property
    def collection_name(self):
        return "licenses"

    def is_valid(self) -> bool:
        return True  # no restrictions on the format

# Cell

@dataclass
class CocoImage(CocoElement):
    id: int
    coco_url: str
    width: Optional[int] = None
    height: Optional[int] = None
    license: Optional[int] = None
    file_name: Optional[str] = None
    flickr_url: Optional[str] = None
    date_captured: Optional[str] = None

    def is_valid(self) -> bool:
        try:
            assert int(self.id) >= 0
            assert self.coco_url
            return True
        except:
            return False

    @property
    def collection_name(self):
        return "images"

    def get_file_name(self) -> str:
        return self.file_name or Path(self.coco_url).name

# Cell

@dataclass
class CocoAnnotation(CocoElement):
    id: int
    image_id: int

    def is_valid(self) -> bool:
        try:
            assert int(self.id) >= 0
            assert int(self.image_id) >= 0
            return True
        except:
            return False

    def get_file_name(self) -> str:
        return f'{self.id}.png'

@dataclass
class CocoObjectDetectionAnnotation(CocoAnnotation):
    category_id: int
    bbox: Tuple[X, Y, W, H]
    supercategory: Optional[str] = None
    area: Optional[int] = None
    iscrowd: Optional[int] = None

    def is_valid(self) -> bool:
        if not super().is_valid():
            return False
        try:
            assert int(self.category_id) >= 0
            x, y, w, h = map(int, self.bbox)
            assert x >= 0, x
            assert y >= 0, y
            assert w >= 0, w
            assert h >= 0, h
            return True
        except:
            return False

# Cell

@dataclass
class CocoCategory(CocoElement):
    id: int

    @abstractmethod
    def get_alias(self):
        raise NotImplementedError

    def is_valid(self) -> bool:
        try:
            assert int(self.id) >= 0
            return True
        except:
            return False

@dataclass
class CocoObjectDetectionCategory(CocoCategory):
    name: str
    supercategory: Optional[str] = None

    def get_dir_name(self):
        name = sanitize_filename(self.name)
        return f'{name}--{self.id}'

    def is_valid(self) -> bool:
        if not super().is_valid():
            return False
        if not self.name:
            return False
        return True

# Cell

@dataclass
class CocoDataset(CocoElement):
    annotations: List[CocoAnnotation] = field(default_factory=list)
    images: List[CocoImage] = field(default_factory=list)
    info: CocoInfo = CocoInfo()
    licenses: List[CocoLicense] = field(default_factory=list)

    @classmethod
    def get_non_collective_elements(cls):
        # TODO: rename to get_individual_elements()
        return ['info']

    @classmethod
    def get_collective_elements(cls):
        default_self = cls()
        non_collective = set(cls.get_non_collective_elements())
        return sorted([f.name for f in fields(default_self) if f.name not in non_collective])

    def to_full_str(self):
        return (
            f'{self.__class__.__name__}(' + \
            ', '.join(f'{k}={len(getattr(self, k))}' for k in self.get_collective_elements()) + \
            ')'
        )

    def is_valid(self) -> bool:
        return all(
            el.is_valid()
            for el in self.annotations + self.images + [self.info] + self.licenses
        )

# Cell

@dataclass
class CocoObjectDetectionDataset(CocoDataset):
    annotations: List[CocoObjectDetectionAnnotation] = field(default_factory=list)
    categories: List[CocoObjectDetectionCategory] = field(default_factory=list)

# Cell

MAP_COCO_TYPE_TO_DATASET_CLASS = {
    "object_detection": CocoObjectDetectionDataset,
}

def get_dataset_class(coco_kind: str):
    try:
        return MAP_COCO_TYPE_TO_DATASET_CLASS[coco_kind]
    except KeyError:
        raise ValueError(f"Not supported dataset kind: {kind}")

# Cell

def merge_datasets(d1: CocoDataset, d2: CocoDataset) -> CocoDataset:
    if d1 is None:
        return d2
    if d2 is None:
        return d1
    assert isinstance(d1, CocoDataset), (type(d1), d1)
    assert isinstance(d2, CocoDataset), (type(d1), d1)
    t1 = type(d1)
    t2 = type(d2)
    assert t1 == t2, f'Cannot merge datasets: {t1} != {t2}'

    D1 = d1.to_dict()
    D2 = d2.to_dict()
    K1 = set(D1.keys())
    K2 = set(D2.keys())
    assert K1 == K2, f'Cannot merge datasets: {K1} != {K2}'

    res = {}
    for k in K1:
        if isinstance(D1[k], list):
            v1 = {x['id']: x for x in (D1[k] or [])}
            v2 = {x['id']: x for x in (D2[k] or [])}
            v_res = {}
            for i in v1:
                if i in v2 and v1[i] != v2[i]:
                    raise ValueError(f'Invalid "{k}" of id={i}: {v1[i]} != {v2[i]}')
                v_res[i] = v1[i]
            for i in v2:
                if i in v1 and v2[i] != v1[i]:
                    raise ValueError(f'Invalid "{k}" of id={i}: {v2[i]} != {v1[i]}')
                v_res[i] = v2[i]
            res[k] = sorted(v_res.values(), key=lambda x: x['id'])
        else:
            v1 = D1[k] or {}
            v2 = D2[k] or {}
            if not v1:
                res[k] = v2
            elif not v2:
                res[k] = v1
            else:
                assert v1 == v2, f'key={k}: unexpectedly: {v1} != {v2}'
                res[k] = v1

    return t1.from_dict(res)

# Cell
def shuffle(arr):
    return random.sample(arr, k=len(arr))

# Cell
def cut_annotations_per_category(coco: CocoDataset, max_annotations_per_category: int) -> CocoDataset:
    """ Returns a copy of the input dataset where each class (category)
        contains up to `max_crops_per_class` crops (annotations)
    """
    imgid2img = {img.id: img for img in coco.images}
    catid2anns = {cat.id: [] for cat in coco.categories}
    for ann in coco.annotations:
        catid2anns[ann.category_id].append(ann)

    images = {}
    annotations = {}
    for _, anns in catid2anns.items():
        if len(anns) > max_annotations_per_category:
            anns = shuffle(anns)[:max_annotations_per_category]
        for ann in anns:
            annotations[ann.id] = ann
            images[ann.image_id] = imgid2img[ann.image_id]
    coco = replace(coco, annotations=sorted(annotations.values(), key=lambda x: x.id))
    coco = replace(coco, images=sorted(images.values(), key=lambda x: x.id))

    return coco

# Cell
def remove_invalid_elements(coco: CocoDataset) -> CocoDataset:
    kw = {
        el: [x for x in getattr(coco, el) if x.is_valid()]
        for el in coco.get_collective_elements()
    }
    res = replace(coco, **kw)

    for el in coco.get_non_collective_elements():
        x = getattr(coco, el)
        if not x.is_valid():
            logger.warning(f"Cannot drop invalid element '{el}' (keeping): {x}")

    return res


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_crop_tree.ipynb (unless otherwise specified).

__all__ = ['load_crop_tree', 'dump_crop_tree']

# Cell

import json
import shutil
import logging
from collections import defaultdict
from dataclasses import dataclass, Field
from typing import *
from pathlib import Path
from multiprocessing import Pool, Lock

from .utils import sort_dict, measure_time, read_image, write_image, cut_bbox
from .coco import *

# Cell

def load_crop_tree(
    source_dir: Union[str, Path],
    base_coco: CocoDataset,
    *,
    kind: str = "object_detection",
) -> CocoDataset:
    """ Load modified set of crops from `{path}/crops` and use it
        to filter out the annotations in `base_coco`.
    """
    dataset_class = get_dataset_class(kind)

    source_dir = Path(source_dir)
    logger.info(f"Loading crop_tree from dir: {source_dir}")
    if not source_dir.is_dir():
        raise ValueError(f"Source dir not found: {source_dir}")

    crops_dir = source_dir / 'crops'
    if not crops_dir.exists():
        raise ValueError(f'Source crops dir not found: {crops_dir}')

    catid2cat = {cat.id: cat for cat in base_coco.categories}
    imgid2img = {img.id: img for img in base_coco.images}
    annid2ann = {ann.id: ann for ann in base_coco.annotations}
    annid2imgid = {ann.id: ann.image_id for ann in base_coco.annotations}

    res_cats = {}
    res_imgs = {}
    res_anns = {}

    with measure_time() as timer1:
        for count1, ann_dir in enumerate(crops_dir.iterdir(), 1):
            cat_id = int(ann_dir.name.split('--')[-1])
            cat = catid2cat[cat_id]

            with measure_time() as timer2:
                for count2, ann_file in enumerate(ann_dir.glob('*.png'), 1):
                    ann_id = int(ann_file.stem)
                    ann = annid2ann[ann_id]
                    img_id = annid2imgid[ann_id]
                    img = imgid2img[img_id]

                    res_cats[cat.id] = cat
                    res_imgs[img.id] = img
                    res_anns[ann.id] = ann
            logger.info(f'  loaded {count2} crops from {ann_dir}: elapsed {timer2.elapsed}')
        logger.info(f'Loaded from {count1} crop directories: elapsed {timer1.elapsed}')

    with measure_time() as timer:
        D = {
            **base_coco.to_dict(),
            'images': list(res_imgs.values()),
            'annotations': list(res_anns.values()),
            'categories': list(res_cats.values()),
        }
    logger.info(f'Dataset dict constructed: elapsed {timer.elapsed}')

    with measure_time() as timer:
        coco = dataset_class.from_dict(D)
    logger.info(f'Dataset object constructed: elapsed {timer.elapsed}: {coco.to_full_str()}')

    return coco

# Cell
def _cut_to_chunks(L: List[Any], n) -> List[List[Any]]:
    assert n > 0
    return [
        L[i: i+n] + [None]*(n - len(L[i: i+n]))
        for i in range(0, len(L), n)
    ]

# Cell

def _process_image(img, anns, images_dir, crops_dir, catid2cat, anns_failed_file):
    file_name = img.get_file_name()
    image_file = images_dir / file_name
    image = None

    for ann in anns:
        cat = catid2cat[ann.category_id]
        cat_dir = crops_dir / cat.get_dir_name()
        cat_dir.mkdir(exist_ok=True)

        ann_file = cat_dir / ann.get_file_name()
        if ann_file.is_file():
            continue
        f
        image = image or read_image(image_file, download_url=img.coco_url)
        box = cut_bbox(image, ann.bbox)
        try:
            write_image(box, ann_file)
        except ValueError as e:
            logger.error(e)
            with anns_failed_file.open('a') as f:
                f.write(json.dumps(ann.to_dict(), ensure_ascii=False) + '\n')


def _process_image_list(l):
    l = l or []
    for (img, anns, images_dir, crops_dir, catid2cat, anns_failed_file) in l:
        _process_image(img, anns, images_dir, crops_dir, catid2cat, anns_failed_file)


def dump_crop_tree(
    coco: CocoDataset,
    target_dir: Union[str, Path],
    *,
    kind: str = 'object_detection',
    skip_nulls: bool = True,
    overwrite: bool = False,
    indent: Optional[int] = 4,
    num_processes: int = 1,
) -> None:
    try:
        from tqdm.auto import tqdm
    except ImportError:
        logger.warning("Could not import tqdm, please run 'pip install tqdm'")
        def tqdm(it, *args, **kwargs):
            yield from it

    dataset_class = get_dataset_class(kind)
    if skip_nulls:
        to_dict_function = dataset_class.to_dict_skip_nulls
    else:
        to_dict_function = dataset_class.to_dict

    target_dir = Path(target_dir)
    logger.info(f"Dumping crop_tree to dir: {target_dir}")

    if overwrite:
        if target_dir.is_dir():
            logger.warning(f'Destination and will be overwritten: {target_dir}')
    elif target_dir.is_dir():
        raise ValueError(f"Destination json tree dir already exists: {target_dir}")

    #if overwrite and target_dir.is_dir():
    #    logger.info(f'Deleting old target directory {target_dir}')
    #    shutil.rmtree(str(target_dir))

    target_dir.mkdir(parents=True, exist_ok=True)
    catid2cat = {cat.id: cat for cat in coco.categories}

    imgid2img = {img.id: img for img in coco.images}
    imgid2anns = defaultdict(list)
    for ann in coco.annotations:
        imgid2anns[ann.image_id].append(ann)

    images_dir = target_dir / 'images'
    images_dir.mkdir(exist_ok=True)

    crops_dir = target_dir / 'crops'
    crops_dir.mkdir(exist_ok=True)

    anns_failed = []
    anns_failed_file = crops_dir / 'crops_failed.ndjson'

    if overwrite and crops_dir.is_dir():
        logger.info(f'Cleaning extra files in root {target_dir}')
        to_remove = []
        all_files = {images_dir} | \
                    {images_dir/img.get_file_name() for img in coco.images} \
                    {crops_dir} | \
                    {crops_dir/catid2cat[cat.id].get_dir_name() for cat in coco.categories} | \
                    {crops_dir/catid2cat[ann.category_id].get_dir_name()/ann.get_file_name() for ann in coco.annotations}
        for p in target_dir.glob('**/*'):
            if p not in all_files:
                to_remove.append(p)
        to_remove = sorted(to_remove)
        removed_str = '\n'.join(map(str, to_remove))
        logger.info(f'Removing {len(to_remove)} files and dirs:\n{removed_str}')
        for p in to_remove:
            if p.is_dir():
                shutil.rmtree(str(p))
            else:
                p.unlink()
        logger.info(f'Removed {len(to_remove)} files and dirs.')

    with measure_time() as timer:
        pairs = [
            (imgid2img[imgid], anns, images_dir, crops_dir, catid2cat, anns_failed_file)
            for (imgid, anns) in imgid2anns.items()
        ]
        chunks = _cut_to_chunks(pairs, num_processes)
        with Pool(num_processes) as pool:
            result = list(tqdm(pool.imap(_process_image_list, chunks), total=len(imgid2anns), desc='Processing images'))
        #process_map(_process_image_list, chunks, total=len(imgid2anns), desc='Processing images', max_workers=num_processes)

    logger.info(f'Crops written to {crops_dir}: elapsed {timer.elapsed}')

    if anns_failed:
        logger.warning(f'Failed to process {len(anns_failed)} crops, see file {anns_failed_file}')